import logging
from datetime import datetime
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional

import typer
from module_qc_data_tools import (
    get_layer_from_sn,
    get_sn_from_connectivity,
    outputDataFrame,
    save_dict_list,
)
from module_qc_database_tools.utils import (
    get_chip_type_from_serial_number,
)

from module_qc_tools.cli.globals import (
    CONTEXT_SETTINGS,
    OPTIONS,
    LogLevel,
)
from module_qc_tools.console import console
from module_qc_tools.measurements.vcal_calibration import (
    run,
)
from module_qc_tools.utils.misc import (
    add_identifiers_metadata,
    copytree,
    get_chip_type,
    load_hw_config,
    load_meas_config,
)
from module_qc_tools.utils.power_supply import power_supply
from module_qc_tools.utils.yarr import yarr

logger = logging.getLogger("measurement")

app = typer.Typer(context_settings=CONTEXT_SETTINGS)

# Taking the test-type from the script name which is the test-code in ProdDB.
TEST_TYPE = Path(__file__).stem


def get_subtest_type(chip_data):
    # Determine subtest type
    chiptype = get_chip_type(chip_data._meta_data["ChipConfigs"])
    monitorV = chip_data._meta_data["ChipConfigs"][chiptype]["GlobalConfig"]["MonitorV"]
    injVcalRange = chip_data._meta_data["ChipConfigs"][chiptype]["GlobalConfig"][
        "InjVcalRange"
    ]

    if monitorV == 7 and injVcalRange == 1:
        return "VCAL_HIGH"
    if monitorV == 7 and injVcalRange == 0:
        return "VCAL_HIGH_SMALL_RANGE"
    if monitorV == 8 and injVcalRange == 1:
        return "VCAL_MED"
    if monitorV == 8 and injVcalRange == 0:
        return "VCAL_MED_SMALL_RANGE"

    return "Unknown"


@app.command()
def main(
    hw_config_path: Path = OPTIONS["config_hw"],
    meas_config_path: Path = OPTIONS["config_meas"],
    base_output_dir: Path = OPTIONS["output_dir"],
    module_connectivity: Optional[Path] = OPTIONS["module_connectivity"],
    _verbosity: LogLevel = OPTIONS["verbosity"],
    perchip: bool = OPTIONS["perchip"],
    use_pixel_config: bool = OPTIONS["use_pixel_config"],
    institution: str = OPTIONS["site"],
    use_calib_adc: bool = OPTIONS["use_calib_ADC"],
    save_local: bool = OPTIONS["save_local"],
    debug_gnd: bool = OPTIONS["debug_gnd"],
):
    """main() creates the qcDataFrame and pass it to the run() where the measurements are stored in the qcDataFrame."""

    timestart = datetime.now().strftime("%Y-%m-%d_%H%M%S")

    # if -o option used, overwrite the default output directory
    output_dir = module_connectivity.parent if module_connectivity else base_output_dir

    if base_output_dir != Path("outputs"):
        output_dir = base_output_dir

    output_dir = output_dir.joinpath("Measurements", TEST_TYPE, timestart)
    # Make output directory and start log file
    output_dir.mkdir(parents=True, exist_ok=True)

    logger.info(f"Start {TEST_TYPE.replace('_',' ')}!")
    logger.info(f"TimeStart: {timestart}")

    logger.addHandler(logging.FileHandler(output_dir.joinpath("output.log")))

    config_hw = load_hw_config(hw_config_path)

    if module_connectivity:
        config_hw["yarr"]["connectivity"] = module_connectivity

    # Taking the module SN from YARR path to config in the connectivity file.
    module_serial = get_sn_from_connectivity(config_hw["yarr"]["connectivity"])
    layer = get_layer_from_sn(module_serial)
    chip_type = get_chip_type_from_serial_number(module_serial)

    # Load the  measurement config and combine it with the hardware config
    config = load_meas_config(
        meas_config_path, test_type=TEST_TYPE, chip_type=chip_type
    )
    config.update(config_hw)

    # initialize hardware
    ps = power_supply(config["power_supply"])
    yr = yarr(config["yarr"])

    if not use_pixel_config:
        yr.omit_pixel_config()

    ps.set(v=config["v_max"], i=config["i_config"][layer])

    try:
        data = run(config, ps, yr, layer, debug_gnd, use_calib_adc)
    except KeyboardInterrupt:
        logger.info("KeyboardInterrupt")
        yr.remove_tmp_connectivity()
    except Exception as err:
        logger.exception(err)
        yr.remove_tmp_connectivity()
        raise typer.Exit(1) from err
    add_identifiers_metadata(data, module_serial, institution)

    alloutput = []
    chipnames = []

    for chip in yr._enabled_chip_positions:
        chiptests = []
        # get name from first subtest dataframe
        chipnames += data[chip][0]._meta_data["Name"]
        for chip_data in data[chip]:
            console.print(data[chip])
            subtest_type = get_subtest_type(chip_data)

            outputDF = outputDataFrame()
            outputDF.set_test_type(TEST_TYPE)
            outputDF.set_subtest_type(subtest_type)
            outputDF.set_results(chip_data)
            chiptests += [outputDF.to_dict()]
        alloutput += chiptests

    with TemporaryDirectory() as tmpdirname:
        if perchip:
            for outputDF, chip_name in zip(alloutput, chipnames):
                save_dict_list(
                    Path(tmpdirname).joinpath(f"{chip_name}.json"),
                    [outputDF],
                )
        else:
            save_dict_list(
                Path(tmpdirname).joinpath(f"{module_serial}.json"),
                alloutput,
            )

        # for now, set to false until localDB upload functionality implemented
        upload_failed = True
        upload_implemented = False

        if not save_local:
            # add in logic here to upload to localDB
            msg = "Not implemented yet"
            raise RuntimeError(msg)

        if upload_failed or save_local:
            copytree(tmpdirname, output_dir)
            if upload_failed and upload_implemented:
                logger.warning(
                    "The upload to localDB failed. Please fix and retry uploading the measurement output again."
                )

            logger.info(f"Writing output measurements in {output_dir}")

    logger.info("Done!")
    logger.info(f"TimeEnd: {datetime.now().strftime('%Y-%m-%d_%H%M%S')}")

    # Delete temporary files
    if not use_pixel_config:
        yr.remove_tmp_connectivity()


if __name__ == "__main__":
    typer.run(main)
